@@ -1,8 +1,8 @@
 #!/bin/sh
 
-export AUTOMAKE_SUFFIX=-1.14.1
+export AUTOMAKE_SUFFIX=-1.16.1
 export AUTOCONF_SUFFIX=-2.69
-export LIBTOOL_SUFFIX=-2.4.2
+export LIBTOOL_SUFFIX=-2.4.6
 
 export ACLOCAL="aclocal${AUTOMAKE_SUFFIX}"
 export AUTOMAKE="automake${AUTOMAKE_SUFFIX}"
@@ -19,5 +19,13 @@ export AUTOM4TE="autom4te${AUTOCONF_SUFFIX}"
 export LIBTOOLIZE="libtoolize${LIBTOOL_SUFFIX}"
 export LIBTOOL="libtool${LIBTOOL_SUFFIX}"
 
-$AUTORECONF -f
+export AUTOGEN=autogen
+
+(cd tests && $AUTOGEN ./Makefile.am.def)
+(cd include && $AUTOGEN ./Makefile.am.def)
+$AUTOGEN ./Makefile.am.def
+$LIBTOOLIZE -vcif
+$ACLOCAL -I m4 --install -Wall --force
+$AUTOMAKE -vcaf
+$AUTOCONF -I m4 -f -Wall
 $AUTOM4TE --language=Autotest -I tests tests/testsuite.at -o tests/testsuite
