@@ -4,7 +4,7 @@
 // Author:  Tad E. Smith
 //
 //
-// Copyright 2001-2015 Tad E. Smith
+// Copyright 2001-2017 Tad E. Smith
 //
 // Licensed under the Apache License, Version 2.0 (the "License");
 // you may not use this file except in compliance with the License.
@@ -34,14 +34,11 @@
 #include <algorithm>
 #include <sstream>
 #include <cstdio>
-#include <cmath>
 #include <stdexcept>
 #include <cmath> // std::fmod
 
-#if defined (__BORLANDC__)
 // For _wrename() and _wremove() on Windows.
-#  include <stdio.h>
-#endif
+#include <stdio.h>
 #include <cerrno>
 #ifdef LOG4CPLUS_HAVE_ERRNO_H
 #include <errno.h>
@@ -168,8 +165,8 @@ rolloverFiles(const tstring& filename, unsigned int maxBackupIndex)
     // Map {(maxBackupIndex - 1), ..., 2, 1} to {maxBackupIndex, ..., 3, 2}
     for (int i = maxBackupIndex - 1; i >= 1; --i)
     {
-        source_oss.str(LOG4CPLUS_TEXT(""));
-        target_oss.str(LOG4CPLUS_TEXT(""));
+        source_oss.str(internal::empty_str);
+        target_oss.str(internal::empty_str);
 
         source_oss << filename << LOG4CPLUS_TEXT(".") << i;
         target_oss << filename << LOG4CPLUS_TEXT(".") << (i+1);
@@ -225,7 +222,7 @@ FileAppenderBase::FileAppenderBase(const tstring& filename_,
     , createDirs (createDirs_)
     , reopenDelay(1)
     , bufferSize (0)
-    , buffer (0)
+    , buffer (nullptr)
     , filename(filename_)
     , localeName (LOG4CPLUS_TEXT ("DEFAULT"))
     , fileOpenMode(mode_)
@@ -239,11 +236,12 @@ FileAppenderBase::FileAppenderBase(const Properties& props,
     , createDirs (false)
     , reopenDelay(1)
     , bufferSize (0)
-    , buffer (0)
+    , buffer (nullptr)
 {
     filename = props.getProperty(LOG4CPLUS_TEXT("File"));
     lockFileName = props.getProperty (LOG4CPLUS_TEXT ("LockFile"));
     localeName = props.getProperty (LOG4CPLUS_TEXT ("Locale"), LOG4CPLUS_TEXT ("DEFAULT"));
+
     props.getBool (immediateFlush, LOG4CPLUS_TEXT("ImmediateFlush"));
     props.getBool (createDirs, LOG4CPLUS_TEXT("CreateDirs"));
     props.getInt (reopenDelay, LOG4CPLUS_TEXT("ReopenDelay"));
@@ -252,6 +250,10 @@ FileAppenderBase::FileAppenderBase(const Properties& props,
     bool app = (mode_ & (std::ios_base::app | std::ios_base::ate)) != 0;
     props.getBool (app, LOG4CPLUS_TEXT("Append"));
     fileOpenMode = app ? std::ios::app : std::ios::trunc;
+
+    if (props.getProperty(LOG4CPLUS_TEXT("TextMode"), LOG4CPLUS_TEXT("Text"))
+        == LOG4CPLUS_TEXT("Binary"))
+        fileOpenMode |= std::ios_base::binary;
 }
 
 
@@ -273,9 +275,8 @@ FileAppenderBase::init()
 
     if (bufferSize != 0)
     {
-        delete[] buffer;
-        buffer = new tchar[bufferSize];
-        out.rdbuf ()->pubsetbuf (buffer, bufferSize);
+        buffer.reset (new tchar[bufferSize]);
+        out.rdbuf ()->pubsetbuf (buffer.get (), bufferSize);
     }
 
     helpers::LockFileGuard guard;
@@ -311,8 +312,7 @@ FileAppenderBase::close()
     thread::MutexGuard guard (access_mutex);
 
     out.close();
-    delete[] buffer;
-    buffer = 0;
+    buffer.reset ();
     closed = true;
 }
 
@@ -382,13 +382,13 @@ FileAppenderBase::reopen()
     // When append never failed and the file re-open attempt must
     // be delayed, set the time when reopen should take place.
     if (reopen_time == log4cplus::helpers::Time () && reopenDelay != 0)
-        reopen_time = log4cplus::helpers::Time::gettimeofday()
-            + log4cplus::helpers::Time(reopenDelay);
+        reopen_time = log4cplus::helpers::now ()
+            + helpers::chrono::seconds (reopenDelay);
     else
     {
         // Otherwise, check for end of the delay (or absence of delay)
         // to re-open the file.
-        if (reopen_time <= log4cplus::helpers::Time::gettimeofday()
+        if (reopen_time <= log4cplus::helpers::now ()
             || reopenDelay == 0)
         {
             // Close the current file
@@ -635,7 +635,8 @@ DailyRollingFileAppender::DailyRollingFileAppender(
     bool immediateFlush_, int maxBackupIndex_, bool createDirs_,
     bool rollOnClose_, const tstring& datePattern_)
     : FileAppender(filename_, std::ios_base::app, immediateFlush_, createDirs_)
-    , maxBackupIndex(maxBackupIndex_), rollOnClose(rollOnClose_)
+    , maxBackupIndex(maxBackupIndex_)
+    , rollOnClose(rollOnClose_)
     , datePattern(datePattern_)
 {
     init(schedule_);
@@ -687,21 +688,45 @@ namespace
 
 static
 Time
-round_time (Time const & t, time_t seconds)
+round_time (Time const & t_, helpers::chrono::seconds seconds)
 {
-    return Time (
-        t.getTime ()
-        - static_cast<time_t>(std::fmod (
-                static_cast<double>(t.getTime ()),
-                static_cast<double>(seconds))));
+    time_t const t = helpers::to_time_t (t_);
+    return helpers::from_time_t (
+        t - static_cast<time_t>(std::fmod (
+                static_cast<double>(t),
+                static_cast<double>(seconds.count ()))));
 }
 
 
 static
 Time
-round_time_and_add (Time const & t, Time const & seconds)
+round_time_and_add (Time const & t, helpers::chrono::seconds const & seconds)
 {
-    return round_time (t, seconds.sec ()) + seconds;
+    return round_time (t, seconds) + seconds;
+}
+
+
+static
+std::chrono::seconds
+local_time_offset (Time const & t)
+{
+    tm time_local, time_gmt;
+
+    helpers::localTime (&time_local, t);
+    helpers::gmTime (&time_gmt, t);
+
+    Time t2 = helpers::from_struct_tm (&time_local);
+    Time t3 = helpers::from_struct_tm (&time_gmt);
+
+    return std::chrono::duration_cast<std::chrono::seconds> (t2 - t3);
+}
+
+
+static
+Time
+adjust_for_time_zone (Time const & t, std::chrono::seconds const & tzoffset)
+{
+    return helpers::time_cast (t - tzoffset);
 }
 
 
@@ -712,51 +737,7 @@ void
 DailyRollingFileAppender::init(DailyRollingFileSchedule sch)
 {
     this->schedule = sch;
-
-    Time now = Time::gettimeofday();
-    now.usec(0);
-    struct tm time;
-    now.localtime(&time);
-
-    time.tm_sec = 0;
-    switch (schedule)
-    {
-    case MONTHLY:
-        time.tm_mday = 1;
-        time.tm_hour = 0;
-        time.tm_min = 0;
-        break;
-
-    case WEEKLY:
-        time.tm_mday -= (time.tm_wday % 7);
-        time.tm_hour = 0;
-        time.tm_min = 0;
-        break;
-
-    case DAILY:
-        time.tm_hour = 0;
-        time.tm_min = 0;
-        break;
-
-    case TWICE_DAILY:
-        if(time.tm_hour >= 12) {
-            time.tm_hour = 12;
-        }
-        else {
-            time.tm_hour = 0;
-        }
-        time.tm_min = 0;
-        break;
-
-    case HOURLY:
-        time.tm_min = 0;
-        break;
-
-    case MINUTELY:
-        break;
-    };
-    now.setTime(&time);
-
+    Time now = helpers::truncate_fractions (helpers::now ());
     scheduledFilename = getFilename(now);
     nextRolloverTime = calculateNextRolloverTime(now);
 }
@@ -872,7 +853,7 @@ DailyRollingFileAppender::rollover(bool alreadyLocked)
     loglog_opening_result (loglog, out, filename);
 
     // Calculate the next rollover time
-    log4cplus::helpers::Time now = Time::gettimeofday();
+    log4cplus::helpers::Time now = helpers::now ();
     if (now >= nextRolloverTime)
     {
         scheduledFilename = getFilename(now);
@@ -881,60 +862,180 @@ DailyRollingFileAppender::rollover(bool alreadyLocked)
 }
 
 
-
+static
 Time
-DailyRollingFileAppender::calculateNextRolloverTime(const Time& t) const
+calculateNextRolloverTime(const Time& t, DailyRollingFileSchedule schedule)
 {
+    namespace chrono = helpers::chrono;
+
+    struct tm next;
     switch(schedule)
     {
     case MONTHLY:
     {
-        struct tm nextMonthTime;
-        t.localtime(&nextMonthTime);
-        nextMonthTime.tm_mon += 1;
-        nextMonthTime.tm_isdst = 0;
+        helpers::localTime (&next, t);
+        next.tm_mon += 1;
+        next.tm_mday = 1; // Round up to next month start
+        next.tm_hour = 0;
+        next.tm_min = 0;
+        next.tm_sec = 0;
+        next.tm_isdst = -1;
 
         Time ret;
-        if(ret.setTime(&nextMonthTime) == -1) {
+        try
+        {
+            ret = helpers::from_struct_tm (&next);
+        }
+        catch (std::runtime_error const & e)
+        {
             helpers::getLogLog().error(
-                LOG4CPLUS_TEXT("DailyRollingFileAppender::calculateNextRolloverTime()-")
-                LOG4CPLUS_TEXT(" setTime() returned error"));
+                LOG4CPLUS_TEXT("calculateNextRolloverTime()-")
+                LOG4CPLUS_TEXT(" from_struct_tm() returned error: ")
+                + LOG4CPLUS_C_STR_TO_TSTRING (e.what ()));
             // Set next rollover to 31 days in future.
-            ret = round_time (t, 24 * 60 * 60) + Time(2678400);
+            ret = round_time (t, chrono::seconds (24 * 60 * 60))
+                + chrono::seconds (2678400);
+        }
+        return ret;
+    }
+
+    case WEEKLY:
+    {
+        helpers::localTime (&next, t);
+        // Round up to next week
+        next.tm_mday += (7 - next.tm_wday + 1);
+        next.tm_hour = 0;
+        next.tm_min = 0;
+        next.tm_sec = 0;
+        next.tm_isdst = -1;
+
+        Time ret;
+        try
+        {
+            ret = helpers::from_struct_tm (&next);
+        }
+        catch (std::runtime_error const & e)
+        {
+            helpers::getLogLog().error(
+                LOG4CPLUS_TEXT("calculateNextRolloverTime()-")
+                LOG4CPLUS_TEXT(" from_struct_tm() returned error: ")
+                + LOG4CPLUS_C_STR_TO_TSTRING (e.what ()));
+            // Set next rollover to 7 days in future.
+            ret = round_time (t, chrono::seconds (24 * 60 * 60))
+                + chrono::seconds (7 * 24 * 60 * 60);
+        }
+        return ret;
+    }
+
+    default:
+        helpers::getLogLog ().error (
+            LOG4CPLUS_TEXT ("calculateNextRolloverTime()-")
+            LOG4CPLUS_TEXT (" unhandled or invalid schedule value"));
+        // Fall through.
+
+    case DAILY:
+    {
+        helpers::localTime(&next, t);
+        next.tm_mday += 1;
+        next.tm_hour = 0;
+        next.tm_min = 0;
+        next.tm_sec = 0;
+        next.tm_isdst = -1;
+
+        Time ret;
+        try
+        {
+            ret = helpers::from_struct_tm (&next);
+        }
+        catch (std::runtime_error const & e)
+        {
+            helpers::getLogLog().error(
+                LOG4CPLUS_TEXT("calculateNextRolloverTime()-")
+                LOG4CPLUS_TEXT(" from_struct_tm() returned error: ")
+                + LOG4CPLUS_C_STR_TO_TSTRING (e.what ()));
+            // Set next rollover to 24 hours in future.
+            ret = round_time (t, chrono::seconds (60 * 60))
+                + chrono::seconds (24 * 60 * 60);
         }
 
         return ret;
     }
 
-    case WEEKLY:
-        return round_time (t, 24 * 60 * 60) + Time (7 * 24 * 60 * 60);
-
-    default:
-        helpers::getLogLog ().error (
-            LOG4CPLUS_TEXT ("DailyRollingFileAppender::calculateNextRolloverTime()-")
-            LOG4CPLUS_TEXT (" invalid schedule value"));
-        // Fall through.
-
-    case DAILY:
-        return round_time_and_add (t, Time (24 * 60 * 60));
-
     case TWICE_DAILY:
-        return round_time_and_add (t, Time (12 * 60 * 60));
+    {
+        helpers::localTime(&next, t);
+        if (next.tm_hour < 12)
+            next.tm_hour = 12;
+        else
+            next.tm_hour = 24;
+        next.tm_min = 0;
+        next.tm_sec = 0;
+        next.tm_isdst = -1;
+
+        Time ret;
+        try
+        {
+            ret = helpers::from_struct_tm (&next);
+        }
+        catch (std::runtime_error const & e)
+        {
+            helpers::getLogLog().error(
+                LOG4CPLUS_TEXT("calculateNextRolloverTime()-")
+                LOG4CPLUS_TEXT(" from_struct_tm() returned error: ")
+                + LOG4CPLUS_C_STR_TO_TSTRING (e.what ()));
+            // Set next rollover to 12 hours in future.
+            ret = round_time (t, chrono::seconds (60 * 60))
+                + chrono::seconds (12 * 60 * 60);
+        }
+
+        return ret;
+    }
 
     case HOURLY:
-        return round_time_and_add (t, Time (60 * 60));
+    {
+        helpers::localTime(&next, t);
+        next.tm_hour += 1;
+        next.tm_min = 0;
+        next.tm_sec = 0;
+        next.tm_isdst = -1;
+
+        Time ret;
+        try
+        {
+            ret = helpers::from_struct_tm (&next);
+        }
+        catch (std::runtime_error const & e)
+        {
+            helpers::getLogLog().error(
+                LOG4CPLUS_TEXT("calculateNextRolloverTime()-")
+                LOG4CPLUS_TEXT(" from_struct_tm() returned error: ")
+                + LOG4CPLUS_C_STR_TO_TSTRING (e.what ()));
+            // Set next rollover to 60 minutes in future.
+            ret = round_time (t, chrono::seconds (60 * 60))
+                + chrono::seconds (60 * 60);
+        }
+
+        return ret;
+    }
 
     case MINUTELY:
-        return round_time_and_add (t, Time (60));
+        return round_time_and_add (t, chrono::seconds (60));
     };
 }
 
 
+Time
+DailyRollingFileAppender::calculateNextRolloverTime(const Time& t) const
+{
+    return helpers::truncate_fractions (
+        log4cplus::calculateNextRolloverTime (t, schedule));
+}
+
 
 tstring
 DailyRollingFileAppender::getFilename(const Time& t) const
 {
-    tchar const * pattern = 0;
+    tchar const * pattern = nullptr;
     if (datePattern.empty())
     {
         switch (schedule)
@@ -975,7 +1076,7 @@ DailyRollingFileAppender::getFilename(const Time& t) const
 
     tstring result (filename);
     result += LOG4CPLUS_TEXT(".");
-    result += t.getFormattedTime(pattern, false);
+    result += helpers::getFormattedTime(pattern, t, false);
     return result;
 }
 
@@ -998,7 +1099,7 @@ preprocessDateTimePattern(const tstring& pattern, DailyRollingFileSchedule& sche
     {
         tchar c = pattern[i];
         size_t end_pos = pattern.find_first_not_of(c, i);
-        int len = (end_pos == tstring::npos ? pattern.length() : end_pos) - i;
+        size_t len = (end_pos == tstring::npos ? pattern.length() : end_pos) - i;
 
         switch (c)
         {
@@ -1156,6 +1257,7 @@ preprocessFilenamePattern(const tstring& pattern, DailyRollingFileSchedule& sche
     return result.str();
 }
 
+
 ///////////////////////////////////////////////////////////////////////////////
 // TimeBasedRollingFileAppender ctors and dtor
 ///////////////////////////////////////////////////////////////////////////////
@@ -1174,7 +1276,10 @@ TimeBasedRollingFileAppender::TimeBasedRollingFileAppender(
     , maxHistory(maxHistory_)
     , cleanHistoryOnStart(cleanHistoryOnStart_)
     , rollOnClose(rollOnClose_)
-{ }
+{
+    filenamePattern = preprocessFilenamePattern(filenamePattern, schedule);
+    init();
+}
 
 TimeBasedRollingFileAppender::TimeBasedRollingFileAppender(
     const log4cplus::helpers::Properties& properties)
@@ -1214,12 +1319,12 @@ TimeBasedRollingFileAppender::init()
 
     FileAppenderBase::init();
 
-    Time now = Time::gettimeofday();
+    Time now = helpers::now();
     nextRolloverTime = calculateNextRolloverTime(now);
 
     if (cleanHistoryOnStart)
     {
-        clean(now + Time(maxHistory*getRolloverPeriodDuration()));
+        clean(now + maxHistory*getRolloverPeriodDuration());
     }
 
     lastHeartBeat = now;
@@ -1238,7 +1343,7 @@ TimeBasedRollingFileAppender::append(const spi::InternalLoggingEvent& event)
 void
 TimeBasedRollingFileAppender::open(std::ios_base::openmode mode)
 {
-    scheduledFilename = Time::gettimeofday().getFormattedTime(filenamePattern, false);
+    scheduledFilename = helpers::getFormattedTime(filenamePattern, helpers::now(), false);
     tstring currentFilename = filename.empty() ? scheduledFilename : filename;
 
     if (createDirs)
@@ -1304,7 +1409,7 @@ TimeBasedRollingFileAppender::rollover(bool alreadyLocked)
         loglog_renaming_result (loglog, filename, scheduledFilename, ret);
     }
 
-    Time now = Time::gettimeofday();
+    Time now = helpers::now();
     clean(now);
 
     open(std::ios::out | std::ios::trunc);
@@ -1315,22 +1420,21 @@ TimeBasedRollingFileAppender::rollover(bool alreadyLocked)
 void
 TimeBasedRollingFileAppender::clean(Time time)
 {
-    Time interval = Time(31*24*3600); // ~1 month
-    if (lastHeartBeat.sec() != 0)
+    Time::duration interval = std::chrono::hours{31*24}; // ~1 month
+    if (lastHeartBeat != Time{})
     {
-        interval = time - lastHeartBeat;
+        interval = time - lastHeartBeat + std::chrono::seconds{1};
     }
-    interval += Time(1);
 
-    int periodDuration = getRolloverPeriodDuration();
-    long periods = static_cast<long>(interval.sec() / periodDuration);
+    Time::duration period = getRolloverPeriodDuration();
+    long periods = long(interval.count () / period.count ());
 
     helpers::LogLog & loglog = helpers::getLogLog();
     for (long i = 0; i < periods; i++)
     {
         long periodToRemove = (-maxHistory - 1) - i;
-        Time timeToRemove = time + Time(periodToRemove * periodDuration);
-        tstring filenameToRemove = timeToRemove.getFormattedTime(filenamePattern, false);
+        Time timeToRemove = time + periodToRemove * period;
+        tstring filenameToRemove = helpers::getFormattedTime(filenamePattern, timeToRemove, false);
         loglog.debug(LOG4CPLUS_TEXT("Removing file ") + filenameToRemove);
         file_remove(filenameToRemove);
     }
@@ -1338,79 +1442,34 @@ TimeBasedRollingFileAppender::clean(Time time)
     lastHeartBeat = time;
 }
 
-int
+Time::duration
 TimeBasedRollingFileAppender::getRolloverPeriodDuration() const
 {
     switch (schedule)
     {
     case MONTHLY:
-        return 31*24*3600;
+        return std::chrono::hours{31*24};
     case WEEKLY:
-        return 7*24*3600;
+        return std::chrono::hours{7*24};
     default:
         helpers::getLogLog ().error (
             LOG4CPLUS_TEXT ("TimeBasedRollingFileAppender::getRolloverPeriodDuration()-")
             LOG4CPLUS_TEXT (" invalid schedule value"));
         // Fall through.
     case DAILY:
-        return 24*3600;
+        return std::chrono::hours{24};
     case HOURLY:
-        return 3600;
+        return std::chrono::hours{1};
     case MINUTELY:
-        return 60;
+        return std::chrono::minutes{1};
     }
 }
 
 Time
 TimeBasedRollingFileAppender::calculateNextRolloverTime(const Time& t) const
 {
-    Time result;
-    struct tm next;
-
-    switch(schedule)
-    {
-    case MONTHLY:
-        t.localtime(&next);
-        next.tm_mon += 1;
-        next.tm_mday = 0; // Round up to next month start
-        next.tm_hour = 0;
-        next.tm_min = 0;
-        next.tm_sec = 0;
-        next.tm_isdst = 0;
-        if (result.setTime(&next) == -1) {
-            result = t + Time(getRolloverPeriodDuration());
-        }
-        break;
-
-    case WEEKLY:
-        t.localtime(&next);
-        next.tm_mday += (7 - next.tm_wday + 1); // Round up to next week
-        next.tm_hour = 0; // Round up to next week start
-        next.tm_min = 0;
-        next.tm_sec = 0;
-        next.tm_isdst = 0;
-        if (result.setTime(&next) == -1) {
-            result = t + Time(getRolloverPeriodDuration());
-        }
-        break;
-
-    default:
-    case DAILY:
-    case HOURLY:
-    case MINUTELY:
-    {
-        int periodDuration = getRolloverPeriodDuration();
-        result = t + Time(periodDuration);
-        time_t seconds = result.sec();
-        int remainder = seconds % periodDuration;
-        result.sec(seconds - remainder);
-        break;
-    }
-    };
-
-    result.usec(0);
-
-    return result;
+    return helpers::truncate_fractions (
+        log4cplus::calculateNextRolloverTime (t, schedule));
 }
 
 } // namespace log4cplus
