@@ -4,7 +4,7 @@
 // Author:  Tad E. Smith
 //
 //
-// Copyright 2003-2015 Tad E. Smith
+// Copyright 2003-2017 Tad E. Smith
 //
 // Licensed under the Apache License, Version 2.0 (the "License");
 // you may not use this file except in compliance with the License.
@@ -19,17 +19,30 @@
 // limitations under the License.
 
 #include <log4cplus/config.hxx>
+
+#if defined (LOG4CPLUS_WITH_UNIT_TESTS)
+// Include full Windows.h early for Catch.
+#  include <log4cplus/config/windowsh-inc-full.h>
+#  define CATCH_CONFIG_RUNNER
+#  include <catch.hpp>
+#endif
+
+#include <log4cplus/initializer.h>
 #include <log4cplus/config/windowsh-inc.h>
 #include <log4cplus/logger.h>
 #include <log4cplus/ndc.h>
 #include <log4cplus/mdc.h>
 #include <log4cplus/helpers/loglog.h>
+#include <log4cplus/internal/customloglevelmanager.h>
 #include <log4cplus/internal/internal.h>
 #include <log4cplus/thread/impl/tls.h>
 #include <log4cplus/thread/syncprims-pub-impl.h>
 #include <log4cplus/helpers/loglog.h>
 #include <log4cplus/spi/factory.h>
 #include <log4cplus/hierarchy.h>
+#if ! defined (LOG4CPLUS_SINGLE_THREADED)
+#include "ThreadPool.h"
+#endif
 #include <cstdio>
 #include <iostream>
 #include <stdexcept>
@@ -50,9 +63,84 @@ LOG4CPLUS_EXPORT tostream & tcerr = std::cerr;
 #endif // UNICODE
 
 
+struct InitializerImpl
+{
+#if ! defined (LOG4CPLUS_SINGLE_THREADED)
+    std::mutex mtx;
+
+    static std::once_flag flag;
+#endif
+
+    unsigned count = 0;
+
+    static InitializerImpl * instance;
+};
+
+#if ! defined (LOG4CPLUS_SINGLE_THREADED)
+std::once_flag InitializerImpl::flag;
+#endif
+InitializerImpl * InitializerImpl::instance;
+
+
+Initializer::Initializer ()
+{
+#if ! defined (LOG4CPLUS_SINGLE_THREADED)
+        std::call_once (InitializerImpl::flag,
+            [&] {
+                InitializerImpl::instance = new InitializerImpl;
+            });
+#else
+        InitializerImpl::instance = new InitializerImpl;
+#endif
+
+    LOG4CPLUS_THREADED (
+        std::unique_lock<std::mutex> guard (
+            InitializerImpl::instance->mtx));
+    if (InitializerImpl::instance->count == 0)
+        initialize ();
+
+    ++InitializerImpl::instance->count;
+}
+
+
+// Forward declaration. Defined in this file.
+void shutdownThreadPool();
+
+Initializer::~Initializer ()
+{
+    bool destroy = false;
+    {
+        LOG4CPLUS_THREADED (
+            std::unique_lock<std::mutex> guard (
+                InitializerImpl::instance->mtx));
+        --InitializerImpl::instance->count;
+        if (InitializerImpl::instance->count == 0)
+        {
+            destroy = true;
+            deinitialize ();
+        }
+    }
+    if (destroy)
+    {
+        delete InitializerImpl::instance;
+        InitializerImpl::instance = nullptr;
+    }
+}
+
+
 namespace
 {
 
+#if ! defined (LOG4CPLUS_SINGLE_THREADED)
+static
+std::unique_ptr<progschj::ThreadPool>
+instantiate_thread_pool ()
+{
+    log4cplus::thread::SignalsBlocker sb;
+    return std::unique_ptr<progschj::ThreadPool>(new progschj::ThreadPool (4));
+}
+#endif
+
 
 //! Default context.
 struct DefaultContext
@@ -60,14 +148,18 @@ struct DefaultContext
     log4cplus::thread::Mutex console_mutex;
     helpers::LogLog loglog;
     LogLevelManager log_level_manager;
+    internal::CustomLogLevelManager custom_log_level_manager;
     helpers::Time TTCCLayout_time_base;
     NDC ndc;
     MDC mdc;
-    Hierarchy hierarchy;
     spi::AppenderFactoryRegistry appender_factory_registry;
     spi::LayoutFactoryRegistry layout_factory_registry;
     spi::FilterFactoryRegistry filter_factory_registry;
     spi::LocaleFactoryRegistry locale_factory_registry;
+#if ! defined (LOG4CPLUS_SINGLE_THREADED)
+    std::unique_ptr<progschj::ThreadPool> thread_pool {instantiate_thread_pool ()};
+#endif
+    Hierarchy hierarchy;
 };
 
 
@@ -79,8 +171,8 @@ enum DCState
 };
 
 
-static DCState default_context_state;
-static DefaultContext * default_context;
+static DCState default_context_state = DC_UNINITIALIZED;
+static DefaultContext * default_context = nullptr;
 
 
 struct destroy_default_context
@@ -88,7 +180,7 @@ struct destroy_default_context
     ~destroy_default_context ()
     {
         delete default_context;
-        default_context = 0;
+        default_context = nullptr;
         default_context_state = DC_DESTROYED;
     }
 } static destroy_default_context_
@@ -123,10 +215,28 @@ alloc_dc ()
 
 static
 DefaultContext *
-get_dc (bool alloc = true)
+get_dc(
+#ifdef LOG4CPLUS_REQUIRE_EXPLICIT_INITIALIZATION
+  bool alloc = false
+#else
+  bool alloc = true
+#endif
+)
 {
-    if (LOG4CPLUS_UNLIKELY (! default_context && alloc))
-        alloc_dc ();
+    if (LOG4CPLUS_UNLIKELY(!default_context))
+    {
+        if (alloc)
+        {
+            alloc_dc();
+        }
+        else
+        {
+#ifdef LOG4CPLUS_REQUIRE_EXPLICIT_INITIALIZATION
+            throw std::logic_error("log4cplus is not initialized"
+                " and implicit initialization is turned off");
+#endif
+        }
+    }
     return default_context;
 }
 
@@ -134,6 +244,16 @@ get_dc (bool alloc = true)
 } // namespace
 
 
+namespace internal {
+
+CustomLogLevelManager & getCustomLogLevelManager ()
+{
+    return get_dc ()->custom_log_level_manager;
+}
+
+} // namespace internal
+
+
 namespace helpers
 {
 
@@ -190,6 +310,47 @@ getMDC ()
 }
 
 
+#if ! defined (LOG4CPLUS_SINGLE_THREADED)
+void
+enqueueAsyncDoAppend (SharedAppenderPtr const & appender,
+    spi::InternalLoggingEvent const & event)
+{
+    get_dc ()->thread_pool->enqueue (
+        [=] ()
+        {
+            appender->asyncDoAppend (event);
+        });
+}
+
+#endif
+
+void
+shutdownThreadPool ()
+{
+#if ! defined (LOG4CPLUS_SINGLE_THREADED)
+    DefaultContext * const dc = get_dc (false);
+    if (dc && dc->thread_pool)
+    {
+        dc->thread_pool.reset ();
+    }
+#endif
+}
+
+
+void
+waitUntilEmptyThreadPoolQueue ()
+{
+#if ! defined (LOG4CPLUS_SINGLE_THREADED)
+    DefaultContext * const dc = get_dc (false);
+    if (dc && dc->thread_pool)
+    {
+        dc->thread_pool->wait_until_empty ();
+        dc->thread_pool->wait_until_nothing_in_flight ();
+    }
+#endif
+}
+
+
 namespace spi
 {
 
@@ -249,7 +410,7 @@ appender_sratch_pad::~appender_sratch_pad ()
 
 
 per_thread_data::per_thread_data ()
-    : fnull (0)
+    : fnull (nullptr)
 { }
 
 
@@ -266,13 +427,13 @@ log4cplus::thread::impl::tls_key_type tls_storage_key;
 #if ! defined (LOG4CPLUS_SINGLE_THREADED) \
     && defined (LOG4CPLUS_THREAD_LOCAL_VAR)
 
-LOG4CPLUS_THREAD_LOCAL_VAR per_thread_data * ptd = 0;
+LOG4CPLUS_THREAD_LOCAL_VAR per_thread_data * ptd = nullptr;
 
 
 per_thread_data *
 alloc_ptd ()
 {
-    per_thread_data * tmp = new per_thread_data;
+    auto * tmp = new per_thread_data;
     set_ptd (tmp);
     // This is a special hack. We set the keys' value to non-NULL to
     // get the ptd_cleanup_func to execute when this thread ends. The
@@ -289,23 +450,27 @@ alloc_ptd ()
 per_thread_data *
 alloc_ptd ()
 {
-    per_thread_data * tmp = new per_thread_data;
+    auto * tmp = new per_thread_data;
     set_ptd (tmp);
     return tmp;
 }
 
 #  endif
 
-
 } // namespace internal
 
 
-void initializeFactoryRegistry();
+void initializeFactoryRegistry ();
 
 
 //! Thread local storage clean up function for POSIX threads.
+#if defined (LOG4CPLUS_USE_WIN32_THREADS)
+static
+void NTAPI
+#else
 static
 void
+#endif
 ptd_cleanup_func (void * arg)
 {
     internal::per_thread_data * const arg_ptd
@@ -325,7 +490,7 @@ ptd_cleanup_func (void * arg)
         // similar constructs with POSIX threads.  Otherwise POSIX
         // calls this cleanup routine more than once if the value
         // stays non-NULL after it returns.
-        thread::impl::tls_set_value (internal::tls_storage_key, 0);
+        thread::impl::tls_set_value (internal::tls_storage_key, nullptr);
     else if (arg)
     {
         // Instead of using internal::get_ptd(false) here we are using
@@ -337,7 +502,7 @@ ptd_cleanup_func (void * arg)
         // unless the value is changed (after the destructor starts)
         // by a call to pthread_setspecific().
         delete arg_ptd;
-        thread::impl::tls_set_value (internal::tls_storage_key, 0);
+        thread::impl::tls_set_value (internal::tls_storage_key, nullptr);
     }
     else
     {
@@ -369,7 +534,7 @@ initializeLog4cplus()
     threadSetup ();
 
     DefaultContext * dc = get_dc (true);
-    dc->TTCCLayout_time_base = helpers::Time::gettimeofday ();
+    dc->TTCCLayout_time_base = helpers::now ();
     Logger::getRoot();
     initializeFactoryRegistry();
 
@@ -384,13 +549,71 @@ initialize ()
 }
 
 
+void
+deinitialize ()
+{
+    shutdownThreadPool();
+    Logger::shutdown ();
+}
+
+
 void
 threadCleanup ()
 {
-    // Do thread-specific cleanup.
-    internal::per_thread_data * ptd = internal::get_ptd (false);
-    delete ptd;
-    internal::set_ptd (0);
+    // Here we check that we can get CRT's heap handle because if we do not
+    // then the following `delete` will fail with access violation in
+    // `RtlFreeHeap()`.
+    //
+    // How is it possible that the CRT heap handle is NULL?
+    //
+    // This function can be called from TLS initializer/terminator by loader
+    // when log4cplus is compiled and linked to as a static library. In case of
+    // other threads temination, it should do its job and free per-thread
+    // data. However, when the whole process is being terminated, it is called
+    // after the CRT has been uninitialized and the CRT heap is not available
+    // any more. In such case, instead of crashing, we just give up and leak
+    // the memory for the short while before the process terminates anyway.
+    //
+    // It is possible to work around this situation in user application by
+    // calling `threadCleanup()` manually before `main()` exits.
+#if defined (_WIN32)
+    if (_get_heap_handle() != 0)
+    {
+#endif
+        // Do thread-specific cleanup.
+        internal::per_thread_data * ptd = internal::get_ptd (false);
+        delete ptd;
+#if defined (_WIN32)
+    }
+    else
+    {
+        OutputDebugString (
+            LOG4CPLUS_TEXT ("log4cplus: ")
+            LOG4CPLUS_TEXT ("CRT heap is already gone in threadCleanup()\n"));
+    }
+#endif
+    internal::set_ptd (nullptr);
+}
+
+
+void
+setThreadPoolSize (std::size_t LOG4CPLUS_THREADED (pool_size))
+{
+#if ! defined (LOG4CPLUS_SINGLE_THREADED)
+    get_dc ()->thread_pool->set_pool_size (pool_size);
+#endif
+}
+
+
+static
+void
+freeTLSSlot ()
+{
+    if (internal::tls_storage_key != thread::impl::tls_key_type ())
+    {
+        thread::impl::tls_cleanup(internal::tls_storage_key);
+        internal::tls_storage_key = thread::impl::tls_key_type();
+    }
 }
 
 
@@ -408,9 +631,11 @@ static
 void
 queueLog4cplusInitializationThroughAPC ()
 {
+#if defined (LOG4CPLUS_BUILD_DLL)
     if (! QueueUserAPC (initializeLog4cplusApcProc, GetCurrentThread (),
         0))
         throw std::runtime_error ("QueueUserAPC() has failed");
+#endif
 }
 
 
@@ -423,11 +648,13 @@ thread_callback (LPVOID /*hinstDLL*/, DWORD fdwReason, LPVOID /*lpReserved*/)
     {
     case DLL_PROCESS_ATTACH:
     {
+#if !defined(LOG4CPLUS_REQUIRE_EXPLICIT_INITIALIZATION)
         // We cannot initialize log4cplus directly here. This is because
         // DllMain() is called under loader lock. When we are using C++11
         // threads and synchronization primitives then there is a deadlock
         // somewhere in internals of std::mutex::lock().
         queueLog4cplusInitializationThroughAPC ();
+#endif
         break;
     }
 
@@ -453,23 +680,51 @@ thread_callback (LPVOID /*hinstDLL*/, DWORD fdwReason, LPVOID /*lpReserved*/)
 
         // Do thread-specific cleanup.
         log4cplus::threadCleanup ();
-#if ! defined (LOG4CPLUS_THREAD_LOCAL_VAR)
-        log4cplus::thread::impl::tls_cleanup (
-            log4cplus::internal::tls_storage_key);
-#endif
+        log4cplus::freeTLSSlot();
+
         break;
     }
 
     } // switch
 }
 
+
+static
+void NTAPI
+thread_callback_initializer(LPVOID hinstDLL, DWORD fdwReason, LPVOID lpReserved)
+{
+    if (fdwReason == DLL_PROCESS_ATTACH
+        || fdwReason == DLL_THREAD_ATTACH)
+        thread_callback(hinstDLL, fdwReason, lpReserved);
+}
+
+static
+void NTAPI
+thread_callback_terminator(LPVOID hinstDLL, DWORD fdwReason, LPVOID lpReserved)
+{
+    if (fdwReason == DLL_THREAD_DETACH
+        || fdwReason == DLL_PROCESS_DETACH)
+        thread_callback(hinstDLL, fdwReason, lpReserved);
+}
+
 #endif
 
 
+#if defined (LOG4CPLUS_WITH_UNIT_TESTS)
+LOG4CPLUS_EXPORT int unit_tests_main (int argc, char* argv[]);
+int
+unit_tests_main (int argc, char * argv[])
+{
+    return Catch::Session ().run (argc, argv);
+}
+
+#endif // defined (LOG4CPLUS_WITH_UNIT_TESTS)
+
 } // namespace log4cplus
 
 
-#if defined (_WIN32) && defined (LOG4CPLUS_BUILD_DLL) && defined (_DLL)
+#if defined (_WIN32)
+#if defined (LOG4CPLUS_BUILD_DLL) && defined (_DLL)
 extern "C"
 BOOL
 WINAPI
@@ -481,66 +736,85 @@ DllMain (LOG4CPLUS_DLLMAIN_HINSTANCE hinstDLL, DWORD fdwReason,
     return TRUE;  // Successful DLL_PROCESS_ATTACH.
 }
 
-#elif defined (_WIN32) \
-    && defined (_MSC_VER) && _MSC_VER >= 1400 && defined (_DLL)
+#elif defined (_MSC_VER) && _MSC_VER >= 1400 && defined (_DLL)
 extern "C"
 {
 
 // This magic has been pieced together from several sources:
 // - <http://www.nynaeve.net/?p=183>
-// - <http://lists.cs.uiuc.edu/pipermail/cfe-dev/2011-November/018818.html>
+// - <http://lists.llvm.org/pipermail/cfe-dev/2011-November/018818.html>
+// - `internal_shared.h` in CRT source in Visual Studio 2015
 
 #pragma data_seg (push, old_seg)
 #ifdef _WIN64
-#pragma const_seg (".CRT$XLX")
+#pragma const_seg (".CRT$XLY")
 extern const
 #else
-#pragma data_seg (".CRT$XLX")
+#pragma data_seg (".CRT$XLY")
 #endif
-PIMAGE_TLS_CALLBACK log4cplus_p_thread_callback = log4cplus::thread_callback;
+PIMAGE_TLS_CALLBACK log4cplus_p_thread_callback_initializer = log4cplus::thread_callback_initializer;
 #pragma data_seg (pop, old_seg)
+
+#pragma data_seg (push, old_seg)
+#ifdef _WIN64
+#pragma const_seg (".CRT$XLAA")
+extern const
+#else
+#pragma data_seg (".CRT$XLAA")
+#endif
+PIMAGE_TLS_CALLBACK log4cplus_p_thread_callback_terminator = log4cplus::thread_callback_terminator;
+#pragma data_seg (pop, old_seg)
+
 #ifdef _WIN64
 #pragma comment (linker, "/INCLUDE:_tls_used")
-#pragma comment (linker, "/INCLUDE:log4cplus_p_thread_callback")
+#if ! defined (LOG4CPLUS_REQUIRE_EXPLICIT_INITIALIZATION)
+#pragma comment (linker, "/INCLUDE:log4cplus_p_thread_callback_initializer")
+#endif
+#pragma comment (linker, "/INCLUDE:log4cplus_p_thread_callback_terminator")
 #else
 #pragma comment (linker, "/INCLUDE:__tls_used")
-#pragma comment (linker, "/INCLUDE:_log4cplus_p_thread_callback")
+#if ! defined (LOG4CPLUS_REQUIRE_EXPLICIT_INITIALIZATION)
+#pragma comment (linker, "/INCLUDE:_log4cplus_p_thread_callback_initializer")
+#endif
+#pragma comment (linker, "/INCLUDE:_log4cplus_p_thread_callback_terminator")
 #endif
 
 } // extern "C"
 
-#elif defined (_WIN32)
+#endif
+
 namespace {
 
 struct _static_log4cplus_initializer
 {
+#if ! defined (LOG4CPLUS_REQUIRE_EXPLICIT_INITIALIZATION)
     _static_log4cplus_initializer ()
     {
         // It is not possible to reliably call initializeLog4cplus() here
         // when we are using Visual Studio and C++11 threads
         // and synchronization primitives. It would result into a deadlock
         // on loader lock.
-#if ! (defined (LOG4CPLUS_WITH_CXX11_THREADS) && defined (_MSC_VER))
+#if ! defined (_MSC_VER)
         log4cplus::initializeLog4cplus ();
 #endif
     }
+#endif
 
     ~_static_log4cplus_initializer ()
     {
         // Last thread cleanup.
         log4cplus::threadCleanup ();
-#if ! defined (LOG4CPLUS_THREAD_LOCAL_VAR)
-    log4cplus::thread::impl::tls_cleanup (
-        log4cplus::internal::tls_storage_key);
-#endif
+        log4cplus::freeTLSSlot();
     }
 } static initializer;
 
 } // namespace
 
-#else
+
+#else // defined (WIN32)
 namespace {
 
+#if ! defined (LOG4CPLUS_REQUIRE_EXPLICIT_INITIALIZATION)
 static void
 _log4cplus_initializer_func ()
     LOG4CPLUS_CONSTRUCTOR_FUNC (LOG4CPLUS_INIT_PRIORITY_BASE);
@@ -549,21 +823,22 @@ _log4cplus_initializer_func ()
 {
     log4cplus::initializeLog4cplus();
 }
+#endif
 
 struct _static_log4cplus_initializer
 {
+#if ! defined (LOG4CPLUS_REQUIRE_EXPLICIT_INITIALIZATION)
     _static_log4cplus_initializer ()
     {
         log4cplus::initializeLog4cplus();
     }
+#endif
 
     ~_static_log4cplus_initializer ()
     {
         // Last thread cleanup.
         log4cplus::threadCleanup ();
-
-        log4cplus::thread::impl::tls_cleanup (
-            log4cplus::internal::tls_storage_key);
+        log4cplus::freeTLSSlot();
     }
 } static initializer
 LOG4CPLUS_INIT_PRIORITY (LOG4CPLUS_INIT_PRIORITY_BASE);
