@@ -1,4 +1,3 @@
-
 #include <log4cplus/logger.h>
 #include <log4cplus/consoleappender.h>
 #include <log4cplus/layout.h>
@@ -6,8 +5,8 @@
 #include <log4cplus/mdc.h>
 #include <log4cplus/helpers/loglog.h>
 #include <log4cplus/thread/threads.h>
-#include <log4cplus/helpers/sleep.h>
 #include <log4cplus/loggingmacros.h>
+#include <log4cplus/initializer.h>
 #include <iostream>
 #include <string>
 
@@ -19,7 +18,7 @@ int
 main()
 {
     cout << "Entering main()..." << endl;
-    log4cplus::initialize ();
+    log4cplus::Initializer initializer;
     LogLog::getLogLog()->setInternalDebugging(true);
     try {
         SharedObjectPtr<Appender> append_1(new ConsoleAppender());
@@ -29,25 +28,25 @@ main()
             LOG4CPLUS_TEXT ("MDC value"));
         log4cplus::tstring pattern = LOG4CPLUS_TEXT("%d{%m/%d/%y %H:%M:%S,%Q} [%t] %-5p %c{2} %%%x%% - %X{key} - %m [%l]%n");
         //	std::tstring pattern = LOG4CPLUS_TEXT("%d{%c} [%t] %-5p [%.15c{3}] %%%x%% - %m [%l]%n");
-        append_1->setLayout( std::auto_ptr<Layout>(new PatternLayout(pattern)) );
+        append_1->setLayout( std::unique_ptr<Layout>(new PatternLayout(pattern)) );
         Logger::getRoot().addAppender(append_1);
 
         Logger logger = Logger::getInstance(LOG4CPLUS_TEXT("test.a.long_logger_name.c.logger"));
         LOG4CPLUS_DEBUG(logger, "This is the FIRST log message...");
 
-        sleep(1, 0);
+        std::this_thread::sleep_for (std::chrono::seconds (1));
         {
             NDCContextCreator ndc(LOG4CPLUS_TEXT("second"));
             LOG4CPLUS_INFO(logger, "This is the SECOND log message...");
         }
 
-        sleep(1, 0);
+        std::this_thread::sleep_for (std::chrono::seconds (1));
         LOG4CPLUS_WARN(logger, "This is the THIRD log message...");
 
-        sleep(1, 0);
+        std::this_thread::sleep_for (std::chrono::seconds (1));
         LOG4CPLUS_ERROR(logger, "This is the FOURTH log message...");
 
-        sleep(1, 0);
+        std::this_thread::sleep_for (std::chrono::seconds (1));
         LOG4CPLUS_FATAL(logger, "This is the FIFTH log message...");
     }
     catch(...) {
@@ -58,4 +57,3 @@ main()
     cout << "Exiting main()..." << endl;
     return 0;
 }
-
